// Tonokip RepRap firmware rewrite based off of Hydra-mmm firmware.
// Licence: GPL

#ifndef MARLIN_H
#define MARLIN_H

#define  FORCE_INLINE __attribute__((always_inline)) inline

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>


#include "fastio.h"
#include "Configuration.h"
#include "pins.h"

#ifndef AT90USB
#define  HardwareSerial_h 
#endif

#if (ARDUINO >= 100)
# include "Arduino.h"
#else
# include "WProgram.h"
  //Arduino < 1.0.0 does not define this, so we need to do it ourselfs
# define analogInputToDigitalPin(p) ((p) + A0)
#endif

#include "MarlinSerial.h"

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#include "WString.h"

#ifdef AT90USB
  #define MYSERIAL Serial
#else
  #define MYSERIAL MSerial
#endif

#define SERIAL_PROTOCOL(x) (MYSERIAL.print(x))
#define SERIAL_PROTOCOL_F(x,y) (MYSERIAL.print(x,y))
#define SERIAL_PROTOCOLPGM(x) (serialprintPGM(PSTR(x)))
#define SERIAL_PROTOCOLLN(x) (MYSERIAL.print(x),MYSERIAL.write('\n'))
#define SERIAL_PROTOCOLLNPGM(x) (serialprintPGM(PSTR(x)),MYSERIAL.write('\n'))


const char errormagic[] PROGMEM ="Error:";
const char echomagic[] PROGMEM ="echo:";
#define SERIAL_ERROR_START (serialprintPGM(errormagic))
#define SERIAL_ERROR(x) SERIAL_PROTOCOL(x)
#define SERIAL_ERRORPGM(x) SERIAL_PROTOCOLPGM(x)
#define SERIAL_ERRORLN(x) SERIAL_PROTOCOLLN(x)
#define SERIAL_ERRORLNPGM(x) SERIAL_PROTOCOLLNPGM(x)

#define SERIAL_ECHO_START (serialprintPGM(echomagic))
#define SERIAL_ECHO(x) SERIAL_PROTOCOL(x)
#define SERIAL_ECHOPGM(x) SERIAL_PROTOCOLPGM(x)
#define SERIAL_ECHOLN(x) SERIAL_PROTOCOLLN(x)
#define SERIAL_ECHOLNPGM(x) SERIAL_PROTOCOLLNPGM(x)

#define SERIAL_ECHOPAIR(name,value) (serial_echopair_P(PSTR(name),(value)))

void serial_echopair_P(const char *s_P, float v);
void serial_echopair_P(const char *s_P, double v);
void serial_echopair_P(const char *s_P, unsigned long v);

FORCE_INLINE void serialprintPGM(const char *str)
{
  char ch=pgm_read_byte(str);
  while(ch)
  {
    MYSERIAL.write(ch);
    ch=pgm_read_byte(++str);
  }
}


void get_command();
void process_commands();

void manage_inactivity();

#if defined(XenablePin) && XenablePin > -1
  #define  enable_x() WRITE(XenablePin, XenableOn)
  #define Disable_X() WRITE(XenablePin,!XenableOn)
#else
  #define enable_x() ;
  #define DisableX() ;
#endif

#if defined(YenablePin) && YenablePin > -1
  #define  enable_y() WRITE(YenablePin, YenableOn)
  #define Disable_Y() WRITE(YenablePin,!YenableOn)
#else
  #define enable_y() ;
  #define DisableY() ;
#endif

#if defined(ZenablePin) && ZenablePin > -1
    #define  enable_z() WRITE(ZenablePin, ZenableOn)
    #define Disable_Z() WRITE(ZenablePin,!ZenableOn)
#else
  #define enable_z() ;
  #define DisableZ() ;
#endif

#if defined(E0enablePin) && (E0enablePin > -1)
  #define enable_e0() WRITE(E0enablePin, EenableOn)
  #define DisableE0() WRITE(E0enablePin,!EenableOn)
#else
  #define enable_e0()  /* nothing */
  #define DisableE0() /* nothing */
#endif

#if (Nozzles > 1) && defined(E1enablePin) && (E1enablePin > -1)
  #define enable_e1() WRITE(E1enablePin, EenableOn)
  #define DisableE1() WRITE(E1enablePin,!EenableOn)
#else
  #define enable_e1()  /* nothing */
  #define DisableE1() /* nothing */
#endif

#if (Nozzles > 2) && defined(E2enablePin) && (E2enablePin > -1)
  #define enable_e2() WRITE(E2enablePin, EenableOn)
  #define DisableE2() WRITE(E2enablePin,!EenableOn)
#else
  #define enable_e2()  /* nothing */
  #define DisableE2() /* nothing */
#endif

#if (Nozzles > 3) && defined(E3enablePin) && (E3enablePin > -1)
  #define enable_e3() WRITE(E3enablePin, EenableOn)
  #define DisableE3() WRITE(E3enablePin,!EenableOn)
#else
  #define enable_e3()  /* nothing */
  #define DisableE3() /* nothing */
#endif

#if (Nozzles > 4) && defined(E4enablePin) && (E4enablePin > -1)
  #define enable_e4() WRITE(E4enablePin, EenableOn)
  #define DisableE4() WRITE(E4enablePin,!EenableOn)
#else
  #define enable_e4()  /* nothing */
  #define DisableE4() /* nothing */
#endif


enum AxisEnum {X_AXIS=0, Y_AXIS=1, Z_AXIS=2, E_AXIS=3};


void FlushSerialRequestResend();
void ClearToSend();

void get_coordinates();
#ifdef DELTA
void calculate_delta(float cartesian[3]);
extern float delta[3];
#endif
void prepare_move();
void kill();
void Stop();

bool IsStopped();

void enquecommand(const char *cmd); 
void enquecommand_P(const char *cmd); 
void prepare_arc_move(char isclockwise);
void clamp_to_software_endstops(float target[3]);

#ifndef CRITICAL_SECTION_START
  #define CRITICAL_SECTION_START  unsigned char _sreg = SREG; cli();
  #define CRITICAL_SECTION_END    SREG = _sreg;
#endif

extern float Homing_Feedrate[];
extern bool axis_relative_modes[];
extern int feedmultiply;
extern int extrudemultiply; 
extern float current_position[NumAxis] ;
extern float add_homeing[3];
#ifdef DELTA
extern float endstop_adj[3];
#endif
extern float min_pos[3];
extern float max_pos[3];
extern int fanSpeed;
#ifdef BARICUDA
extern int ValvePressure;
extern int EtoPPressure;
#endif

#ifdef FAN_SOFT_PWM
extern unsigned char fanSpeedSoftPwm;
#endif


extern unsigned long starttime;
extern unsigned long stoptime;
extern uint8_t active_extruder;

#endif
