/*
  planner.h - buffers movement commands and manages the acceleration profile plan
  Part of Grbl

  Copyright (c) 2009-2011 Simen Svale Skogsrud

  Grbl is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Grbl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Grbl.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef planner_h
#define planner_h

#include "Marlin.h"

typedef struct {
  long steps_x, steps_y, steps_z, steps_e; 
  unsigned long step_event_count;           
  long accelerate_until;                   
  long decelerate_after;                   
  long acceleration_rate;                   
  unsigned char direction_bits;          
  unsigned char active_extruder;    
  float nominal_speed;                        
  float entry_speed;                                
  float max_entry_speed;                            
  float millimeters;                                
  float acceleration;                               
  unsigned char recalculate_flag;                    
  unsigned char nominal_length_flag;                 
  unsigned long nominal_rate;                      
  unsigned long initial_rate;                      
  unsigned long final_rate;                        
  unsigned long acceleration_st;                   
  unsigned long fan_speed;       
  #ifdef ADVANCE
    long advance_rate;
    volatile long initial_advance;
    volatile long final_advance;
    float advance;
  #endif
  #ifdef BARICUDA
  unsigned long valve_pressure;
  unsigned long e_to_p_pressure;
  #endif
  volatile char busy;
} block_t;


void plan_init();

void plan_buffer_line(const float &x, const float &y, const float &z, const float &e, float feed_rate, const uint8_t &extruder);


void plan_set_position(const float &x, const float &y, const float &z, const float &e);
void plan_set_e_position(const float &e);



void check_axes_activity();
uint8_t movesplanned();

extern unsigned long minsegmenttime;
extern float max_feedrate[4];
extern float axis_steps_per_unit[4];
extern unsigned long max_acceleration_units_per_sq_second[4]; 
extern float minimumfeedrate;
extern float acceleration;       
extern float retract_acceleration; 
extern float max_xy_jerk;
extern float max_z_jerk;
extern float max_e_jerk;
extern float mintravelfeedrate;
extern unsigned long axis_steps_per_sqr_second[NumAxis];

#ifdef AUTOTEMP
    extern bool autotemp_enabled;
    extern float autotemp_max;
    extern float autotemp_min;
    extern float autotemp_factor;
#endif
	
extern block_t block_buffer[BLOCK_BUFFER_SIZE];           
extern volatile unsigned char block_buffer_head;     
extern volatile unsigned char block_buffer_tail; 
  
FORCE_INLINE void plan_discard_current_block()  
{
  if (block_buffer_head != block_buffer_tail) {
    block_buffer_tail = (block_buffer_tail + 1) & (BLOCK_BUFFER_SIZE - 1);  
  }
}

FORCE_INLINE block_t *plan_get_current_block() 
{
  if (block_buffer_head == block_buffer_tail) { 
    return(NULL); 
  }
  block_t *block = &block_buffer[block_buffer_tail];
  block->busy = true;
  return(block);
}

FORCE_INLINE bool blocks_queued() 
{
  if (block_buffer_head == block_buffer_tail) { 
    return false; 
  }
  else
    return true;
}

#ifdef PreventDangerousExtrude
void set_extrude_min_temp(float temp);
#endif

void reset_acceleration_rates();
#endif
