/*
  stepper.h - stepper motor driver: executes motion plans of planner.c using the stepper motors
  Part of Grbl

  Copyright (c) 2009-2011 Simen Svale Skogsrud

  Grbl is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Grbl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Grbl.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef stepper_h
#define stepper_h 

#include "planner.h"

#if Nozzles > 4 
  #define WRITE_E_STEP(v) { if(current_block->active_extruder == 4) { WRITE(E4stepPin, v); } else { if(current_block->active_extruder == 3) { WRITE(E3stepPin, v); } else { if(current_block->active_extruder == 2) { WRITE(E2stepPin, v); } else { if(current_block->active_extruder == 1) { WRITE(E1stepPin, v); } else { WRITE(E0stepPin, v); }}}}}
  #define NORM_E_DIR() { if(current_block->active_extruder == 4) { WRITE(E4dirPin, !InvertE4dir); } else { if(current_block->active_extruder == 3) { WRITE(E3dirPin, !InvertE3dir); } else { if(current_block->active_extruder == 2) { WRITE(E2dirPin, !InvertE2dir); } else { if(current_block->active_extruder == 1) { WRITE(E1dirPin, !InvertE1dir); } else { WRITE(E0dirPin, !InvertE0dir); }}}}}
  #define REV_E_DIR() { if(current_block->active_extruder == 4) { WRITE(E4dirPin, InvertE4dir); } else { if(current_block->active_extruder == 3) { WRITE(E3dirPin, InvertE3dir); } else { if(current_block->active_extruder == 2) { WRITE(E2dirPin, InvertE2dir); } else { if(current_block->active_extruder == 1) { WRITE(E1dirPin, InvertE1dir); } else { WRITE(E0dirPin, InvertE0dir); }}}}}
#elif Nozzles > 3
  #define WRITE_E_STEP(v) { if(current_block->active_extruder == 3) { WRITE(E3stepPin, v); } else { if(current_block->active_extruder == 2) { WRITE(E2stepPin, v); } else { if(current_block->active_extruder == 1) { WRITE(E1stepPin, v); } else { WRITE(E0stepPin, v); }}}}
  #define NORM_E_DIR() { if(current_block->active_extruder == 3) { WRITE(E3dirPin, !InvertE3dir); } else { if(current_block->active_extruder == 2) { WRITE(E2dirPin, !InvertE2dir); } else { if(current_block->active_extruder == 1) { WRITE(E1dirPin, !InvertE1dir); } else { WRITE(E0dirPin, !InvertE0dir); }}}}
  #define REV_E_DIR() { if(current_block->active_extruder == 3) { WRITE(E3dirPin, InvertE3dir); } else { if(current_block->active_extruder == 2) { WRITE(E2dirPin, InvertE2dir); } else { if(current_block->active_extruder == 1) { WRITE(E1dirPin, InvertE1dir); } else { WRITE(E0dirPin, InvertE0dir); }}}}
#elif Nozzles > 2
  #define WRITE_E_STEP(v) { if(current_block->active_extruder == 2) { WRITE(E2stepPin, v); } else { if(current_block->active_extruder == 1) { WRITE(E1stepPin, v); } else { WRITE(E0stepPin, v); }}}
  #define NORM_E_DIR() { if(current_block->active_extruder == 2) { WRITE(E2dirPin, !InvertE2dir); } else { if(current_block->active_extruder == 1) { WRITE(E1dirPin, !InvertE1dir); } else { WRITE(E0dirPin, !InvertE0dir); }}}
  #define REV_E_DIR() { if(current_block->active_extruder == 2) { WRITE(E2dirPin, InvertE2dir); } else { if(current_block->active_extruder == 1) { WRITE(E1dirPin, InvertE1dir); } else{ WRITE(E0dirPin, InvertE0dir); }}}
#elif Nozzles > 1
  #ifndef DUAL_X_CARRIAGE
    #define WRITE_E_STEP(v){if(current_block->active_extruder == 1) { WRITE(E1stepPin, v); } else { WRITE(E0stepPin, v); }}
    #define NORM_E_DIR() { if(current_block->active_extruder == 1) { WRITE(E1dirPin, !InvertE1dir); } else { WRITE(E0dirPin, !InvertE0dir); }}
    #define REV_E_DIR() { if(current_block->active_extruder == 1) { WRITE(E1dirPin, InvertE1dir); } else { WRITE(E0dirPin, InvertE0dir); }}
  #else
    extern bool extruder_duplication_enabled;
    #define WRITE_E_STEP(v) { if(extruder_duplication_enabled) { WRITE(E0stepPin, v); WRITE(E1stepPin, v); } else if(current_block->active_extruder == 1) { WRITE(E1stepPin, v); } else { WRITE(E0stepPin, v); }}
    #define NORM_E_DIR() { if(extruder_duplication_enabled) { WRITE(E0dirPin, !InvertE0dir); WRITE(E1dirPin, !InvertE1dir); } else if(current_block->active_extruder == 1) { WRITE(E1dirPin, !InvertE1dir); } else { WRITE(E0dirPin, !InvertE0dir); }}
    #define REV_E_DIR() { if(extruder_duplication_enabled) { WRITE(E0dirPin, InvertE0dir); WRITE(E1dirPin, InvertE1dir); } else if(current_block->active_extruder == 1) { WRITE(E1dirPin, InvertE1dir); } else { WRITE(E0dirPin, InvertE0dir); }}
  #endif  
#else
  #define WRITE_E_STEP(v) WRITE(E0stepPin, v)
  #define NORM_E_DIR() WRITE(E0dirPin, !InvertE0dir)
  #define REV_E_DIR() WRITE(E0dirPin, InvertE0dir)
#endif

#ifdef ABORT_ON_ENDSTOP_HIT_FEATURE_ENABLED
extern bool abort_on_endstop_hit;
#endif

void st_init();
void st_synchronize();
void st_set_position(const long &x, const long &y, const long &z, const long &e);
void st_set_e_position(const long &e);
long st_get_position(uint8_t axis);
void st_wake_up();
void checkHitEndstops(); 
void endstops_hit_on_purpose(); 
void enable_endstops(bool check);
void checkStepperErrors(); 
void finishAndDisableSteppers();
extern block_t *current_block;
void quickStop();
void digitalPotWrite(int address, int value);
void microstep_ms(uint8_t driver, int8_t ms1, int8_t ms2);
void microstep_mode(uint8_t driver, uint8_t stepping);
void digipot_init();
void digipot_current(uint8_t driver, int current);
void microstep_init();
void microstep_readings();

#endif
