#include "Marlin.h"
#include "temperature.h"

int target_temperature[Nozzles] = { 0 };
int target_temperature_bed = 0;
int target_temperature_LC = 19; 
int current_temperature_raw[Nozzles] = { 0 };
float current_temperature[Nozzles] = { 0.0 };
int current_temperature_LC_raw = 0; 
float current_temperature_LC = 0.0;
int current_temperature_bed_raw = 0;
float current_temperature_bed = 0.0;
#ifdef Temp2_AS_REDUNDANT
  int redundant_temperature_raw = 0;
  float redundant_temperature = 0.0;
#endif
#ifdef PidTemp
  float Kp=DefaultKp;
  float Ki=(DefaultKi*Pid_dT);
  float Kd=(DefaultKd/Pid_dT);
  #ifdef PID_ADD_EXTRUSION_RATE
    float Kc=DEFAULT_Kc;
  #endif
#endif

#ifdef PidTempBED
  float bedKp=DefaultBedKp;
  float bedKi=(DefaultBedKi*Pid_dT);
  float bedKd=(DefaultBedKd/Pid_dT);
#endif
  
#ifdef FAN_SOFT_PWM
  unsigned char fanSpeedSoftPwm;
#endif
  
static volatile bool temp_meas_ready = false;

#ifdef PidTemp
  static float temp_iState[Nozzles] = { 0 };
  static float temp_dState[Nozzles] = { 0 };
  static float pTerm[Nozzles];
  static float iTerm[Nozzles];
  static float dTerm[Nozzles];
  static float pid_error[Nozzles];
  static float temp_iState_min[Nozzles];
  static float temp_iState_max[Nozzles];
  static bool pid_reset[Nozzles];
#endif
#ifdef PidTempBED
  static float temp_iState_bed = { 0 };
  static float temp_dState_bed = { 0 };
  static float pTerm_bed;
  static float iTerm_bed;
  static float dTerm_bed;
  static float pid_error_bed;
  static float temp_iState_min_bed;
  static float temp_iState_max_bed;
#else
	static unsigned long  previous_millis_bed_heater;
#endif
  static unsigned char soft_pwm[Nozzles];
  static unsigned char soft_pwm_bed;
#ifdef FAN_SOFT_PWM
  static unsigned char soft_pwm_fan;
#endif
#if (defined(EXTRUDER_0_AUTO_FanPin) && EXTRUDER_0_AUTO_FanPin > -1) || \
    (defined(EXTRUDER_1_AUTO_FanPin) && EXTRUDER_1_AUTO_FanPin > -1) || \
    (defined(EXTRUDER_2_AUTO_FanPin) && EXTRUDER_2_AUTO_FanPin > -1)
  static unsigned long extruder_autofan_last_check;
#endif  

#if Nozzles > 5
  # error Unsupported number of Nozzles
#elif Nozzles > 4
  # define ARRAY_BY_Nozzles(v1, v2, v3, v4, v5) { v1, v2, v3, v4, v5 }
#elif Nozzles > 3
  # define ARRAY_BY_Nozzles(v1, v2, v3, v4, v5) { v1, v2, v3, v4 }
#elif Nozzles > 2
  # define ARRAY_BY_Nozzles(v1, v2, v3, v4, v5) { v1, v2, v3 }
#elif Nozzles > 1
  # define ARRAY_BY_Nozzles(v1, v2, v3, v4, v5) { v1, v2 }
#else
  # define ARRAY_BY_Nozzles(v1, v2, v3, v4, v5) { v1 }
#endif

static int minttemp_raw[Nozzles] = ARRAY_BY_Nozzles( HEATER_0_RAW_LO_TEMP , HEATER_1_RAW_LO_TEMP , HEATER_2_RAW_LO_TEMP , HEATER_3_RAW_LO_TEMP , HEATER_4_RAW_LO_TEMP);
static int maxttemp_raw[Nozzles] = ARRAY_BY_Nozzles( HEATER_0_RAW_HI_TEMP , HEATER_1_RAW_HI_TEMP , HEATER_2_RAW_HI_TEMP , HEATER_3_RAW_HI_TEMP , HEATER_4_RAW_HI_TEMP);
static int minttemp[Nozzles] = ARRAY_BY_Nozzles( 0, 0, 0, 0, 0);
static int maxttemp[Nozzles] = ARRAY_BY_Nozzles( 16383, 16383, 16383, 16383, 16383);
#ifdef BedMaxTemp
static int bed_maxttemp_raw = HEATER_BED_RAW_HI_TEMP;
#endif

#ifdef Temp2_AS_REDUNDANT
  static void *heater_ttbl_map[2] = {(void *)HEATER_0_TEMPTABLE, (void *)HEATER_1_TEMPTABLE };
  static uint8_t heater_ttbllen_map[2] = { HEATER_0_TEMPTABLE_LEN, HEATER_1_TEMPTABLE_LEN };
#else
  static void *heater_ttbl_map[Nozzles] = ARRAY_BY_Nozzles( (void *)HEATER_0_TEMPTABLE, (void *)HEATER_1_TEMPTABLE, (void *)HEATER_2_TEMPTABLE, (void *)HEATER_3_TEMPTABLE, (void *)HEATER_4_TEMPTABLE );
  static uint8_t heater_ttbllen_map[Nozzles] = ARRAY_BY_Nozzles( HEATER_0_TEMPTABLE_LEN, HEATER_1_TEMPTABLE_LEN, HEATER_2_TEMPTABLE_LEN, HEATER_3_TEMPTABLE_LEN, HEATER_4_TEMPTABLE_LEN );
#endif

static float analog2temp(int raw, uint8_t e);
static float analog2tempBed(int raw);
static float analog2tempLC(int raw);
static void updateTemperaturesFromRawValues();

#ifdef WATCH_TEMP_PERIOD
int watch_start_temp[Nozzles] = ARRAY_BY_Nozzles(0,0,0,0,0);
unsigned long watchmillis[Nozzles] = ARRAY_BY_Nozzles(0,0,0,0,0);
#endif 

#ifndef SOFT_PWM_SCALE
#define SOFT_PWM_SCALE 0
#endif

void PID_autotune(float temp, int extruder, int ncycles)
{
  float input = 0.0;
  int cycles=0;
  bool heating = true;

  unsigned long temp_millis = millis();
  unsigned long t1=temp_millis;
  unsigned long t2=temp_millis;
  long t_high = 0;
  long t_low = 0;

  long bias, d;
  float Ku, Tu;
  float Kp, Ki, Kd;
  float max = 0, min = 10000;

  if ((extruder > Nozzles)
  #if (Temp3Pin <= -1)
       ||(extruder < 0)
  #endif
       ){
          SERIAL_ECHOLN("PID Autotune failed. Bad extruder number.");
          return;
        }
	
  SERIAL_ECHOLN("PID Autotune start");
  
  disable_heater();

  if (extruder<0)
  {
     soft_pwm_bed = (MaxBedPower)/2;
     bias = d = (MaxBedPower)/2;
   }
   else
   {
     soft_pwm[extruder] = (PidMax)/2;
     bias = d = (PidMax)/2;
  }




 for(;;) {

    if(temp_meas_ready == true) 
	{
      updateTemperaturesFromRawValues();

      input = (extruder<0)?current_temperature_bed:current_temperature[extruder];

      max=max(max,input);
      min=min(min,input);
      if(heating == true && input > temp) 
	  {
        if(millis() - t2 > 5000) 
		{ 
          heating=false;
          if (extruder<0)
            soft_pwm_bed = (bias - d) >> 1;
          else
            soft_pwm[extruder] = (bias - d) >> 1;
          t1=millis();
          t_high=t1 - t2;
          max=temp;
        }
      }
      if(heating == false && input < temp) 
	  {
        if(millis() - t1 > 5000)
		{
          heating=true;
          t2=millis();
          t_low=t2 - t1;
          if(cycles > 0)
		  {
            bias += (d*(t_high - t_low))/(t_low + t_high);
            bias = constrain(bias, 20 ,(extruder<0?(MaxBedPower):(PidMax))-20);
            if(bias > (extruder<0?(MaxBedPower):(PidMax))/2) d = (extruder<0?(MaxBedPower):(PidMax)) - 1 - bias;
            else d = bias;

            SERIAL_PROTOCOLPGM(" bias: "); SERIAL_PROTOCOL(bias);
            SERIAL_PROTOCOLPGM(" d: "); SERIAL_PROTOCOL(d);
            SERIAL_PROTOCOLPGM(" min: "); SERIAL_PROTOCOL(min);
            SERIAL_PROTOCOLPGM(" max: "); SERIAL_PROTOCOLLN(max);
            if(cycles > 2) {
              Ku = (4.0*d)/(3.14159*(max-min)/2.0);
              Tu = ((float)(t_low + t_high)/1000.0);
              SERIAL_PROTOCOLPGM(" Ku: "); SERIAL_PROTOCOL(Ku);
              SERIAL_PROTOCOLPGM(" Tu: "); SERIAL_PROTOCOLLN(Tu);
              Kp = 0.6*Ku;
              Ki = 2*Kp/Tu;
              Kd = Kp*Tu/8;
              SERIAL_PROTOCOLLNPGM(" Clasic PID ");
              SERIAL_PROTOCOLPGM(" Kp: "); SERIAL_PROTOCOLLN(Kp);
              SERIAL_PROTOCOLPGM(" Ki: "); SERIAL_PROTOCOLLN(Ki);
              SERIAL_PROTOCOLPGM(" Kd: "); SERIAL_PROTOCOLLN(Kd);
            }
          }
          if (extruder<0)
            soft_pwm_bed = (bias + d) >> 1;
          else
            soft_pwm[extruder] = (bias + d) >> 1;
          cycles++;
          min=temp;
        }
      } 
    }
    if(input > (temp + 20)) {
      SERIAL_PROTOCOLLNPGM("PID Autotune failed! Temperature too high");
      return;
    }
    if(millis() - temp_millis > 2000)
	{
      int p;
      if (extruder<0)
	  {
        p=soft_pwm_bed;       
        SERIAL_PROTOCOLPGM("ok B:");
      }else{
        p=soft_pwm[extruder];       
        SERIAL_PROTOCOLPGM("ok T:");
      }
			
      SERIAL_PROTOCOL(input);   
      SERIAL_PROTOCOLPGM(" @:");
      SERIAL_PROTOCOLLN(p);       

      temp_millis = millis();
    }
    if(((millis() - t1) + (millis() - t2)) > (10L*60L*1000L*2L)) 
	{
      SERIAL_PROTOCOLLNPGM("PID Autotune failed! timeout");
      return;
    }
    if(cycles > ncycles) {
      SERIAL_PROTOCOLLNPGM("PID Autotune finished! Put the Kp, Ki and Kd constants into Configuration.h");
      return;
    }
  }
}

void updatePID()
{
#ifdef PidTemp
  for(int e = 0; e < Nozzles; e++) 
  { 
     temp_iState_max[e] = PidIntegralDriveMax / Ki;  
  }
#endif
#ifdef PidTempBED
  temp_iState_max_bed = PidIntegralDriveMax / bedKi;  
#endif
}
  
int getHeaterPower(int heater)
{
	if (heater<0)
		return soft_pwm_bed;
  return soft_pwm[heater];
}

void manage_heater()
{
  float pid_input;
  float pid_output;

  if(temp_meas_ready != true)
    return; 

  updateTemperaturesFromRawValues();

  for(int e = 0; e < Nozzles; e++) 
  {


    pid_input = current_temperature[e];

	pid_error[e] = target_temperature[e] - pid_input;
	if(pid_error[e] > 10)
	{
		pid_output = 255;
		pid_reset[e] = true;
	}
	else if(pid_error[e] < -10 || target_temperature[e] == 0) 
	{
		pid_output = 0;
		pid_reset[e] = true;
	}
	else 
	{
		if(pid_reset[e] == true) 
		{
			temp_iState[e] = 0.0;
			pid_reset[e] = false;
		}

		pTerm[e] = Kp * pid_error[e];
		temp_iState[e] += pid_error[e];
		temp_iState[e] = constrain(temp_iState[e], temp_iState_min[e], temp_iState_max[e]);
		iTerm[e] = Ki * temp_iState[e];

		#define K2 (1.0-K1)
		dTerm[e] = (Kd * (pid_input - temp_dState[e]))*K2 + (K1 * dTerm[e]);


		pid_output = constrain(pTerm[e] + iTerm[e] - dTerm[e], 0, PidMax);
	}
	temp_dState[e] = pid_input;

    if((current_temperature[e] > minttemp[e]) && (current_temperature[e] < maxttemp[e])) 
    {
      soft_pwm[e] = (int)pid_output >> 1;
    }
    else 
	{
      soft_pwm[e] = 0;
    }
  }

 

  #if Temp3 != 0
	pid_input = current_temperature_bed;
	pid_error_bed = target_temperature_bed - pid_input;
	pTerm_bed = bedKp * pid_error_bed;
	temp_iState_bed += pid_error_bed;
	temp_iState_bed = constrain(temp_iState_bed, temp_iState_min_bed, temp_iState_max_bed);
	iTerm_bed = bedKi * temp_iState_bed;

	#define K2 (1.0-K1)
	dTerm_bed= (bedKd * (pid_input - temp_dState_bed))*K2 + (K1 * dTerm_bed);
	temp_dState_bed = pid_input;

	pid_output = constrain(pTerm_bed + iTerm_bed - dTerm_bed, 0, MaxBedPower);


	if((current_temperature_bed > Heater6MinTemp) && (current_temperature_bed < BedMaxTemp)) 
	{
		soft_pwm_bed = (int)pid_output >> 1;
	}
	else 
	{
		soft_pwm_bed = 0;
	}
  #endif
}

#define PGM_RD_W(x)   (short)pgm_read_word(&x)
static float analog2temp(int raw, uint8_t e) {
#ifdef Temp2_AS_REDUNDANT
  if(e > Nozzles)
#else
  if(e >= Nozzles)
#endif
  {
      SERIAL_ERROR_START;
      SERIAL_ERROR((int)e);
      SERIAL_ERRORLNPGM(" - Invalid extruder number !");
      kill();
  } 
  #ifdef HEATER_0_USES_MAX6675
    if (e == 0)
    {
      return 0.25 * raw;
    }
  #endif

  if(heater_ttbl_map[e] != NULL)
  {
    float celsius = 0;
    uint8_t i;
    short (*tt)[][2] = (short (*)[][2])(heater_ttbl_map[e]);

    for (i=1; i<heater_ttbllen_map[e]; i++)
    {
      if (PGM_RD_W((*tt)[i][0]) > raw)
      {
        celsius = PGM_RD_W((*tt)[i-1][1]) + 
          (raw - PGM_RD_W((*tt)[i-1][0])) * 
          (float)(PGM_RD_W((*tt)[i][1]) - PGM_RD_W((*tt)[i-1][1])) /
          (float)(PGM_RD_W((*tt)[i][0]) - PGM_RD_W((*tt)[i-1][0]));
        break;
      }
    }

    if (i == heater_ttbllen_map[e]) celsius = PGM_RD_W((*tt)[i-1][1]);
	    return celsius;
  }

  return (((raw + TEMP_SENSOR_AD8497_OFFSET) / 100) * 5.0392);
}

static float analog2tempLC(int raw) {
    float celsius = 0;
    byte i;

    for (i=1; i<LCTEMPTABLE_LEN; i++)
    {
      if (PGM_RD_W(LCTEMPTABLE[i][0]) > raw)
      {
        celsius  = PGM_RD_W(LCTEMPTABLE[i-1][1]) + 
          (raw - PGM_RD_W(LCTEMPTABLE[i-1][0])) * 
          (float)(PGM_RD_W(LCTEMPTABLE[i][1]) - PGM_RD_W(LCTEMPTABLE[i-1][1])) /
          (float)(PGM_RD_W(LCTEMPTABLE[i][0]) - PGM_RD_W(LCTEMPTABLE[i-1][0]));
        break;
      }
    }

    if (i == LCTEMPTABLE_LEN) 
		celsius = PGM_RD_W(LCTEMPTABLE[i-1][1]);

    return celsius;
}

static float analog2tempBed(int raw) 
{
    float celsius = 0;
    byte i;

    for (i=1; i<BEDTEMPTABLE_LEN; i++)
    {
      if (PGM_RD_W(BEDTEMPTABLE[i][0]) > raw)
      {
        celsius  = PGM_RD_W(BEDTEMPTABLE[i-1][1]) + 
          (raw - PGM_RD_W(BEDTEMPTABLE[i-1][0])) * 
          (float)(PGM_RD_W(BEDTEMPTABLE[i][1]) - PGM_RD_W(BEDTEMPTABLE[i-1][1])) /
          (float)(PGM_RD_W(BEDTEMPTABLE[i][0]) - PGM_RD_W(BEDTEMPTABLE[i-1][0]));
        break;
      }
    }
    if (i == BEDTEMPTABLE_LEN) celsius = PGM_RD_W(BEDTEMPTABLE[i-1][1]);

    return celsius;
}

static void updateTemperaturesFromRawValues()
{
    for(uint8_t e=0;e<Nozzles;e++)
    {
        current_temperature[e] = analog2temp(current_temperature_raw[e], e);
    }
	current_temperature_LC = analog2tempLC(current_temperature_LC_raw);
    current_temperature_bed = analog2tempBed(current_temperature_bed_raw);
    CRITICAL_SECTION_START;
    temp_meas_ready = false;
    CRITICAL_SECTION_END;
}

void tp_init()
{
  for(int e = 0; e < Nozzles; e++) 
  {
    maxttemp[e] = maxttemp[0];
	#ifdef PidTemp
		temp_iState_min[e] = 0.0;
		temp_iState_max[e] = PidIntegralDriveMax / Ki;
	#endif
	#ifdef PidTempBED
		temp_iState_min_bed = 0.0;
		temp_iState_max_bed = PidIntegralDriveMax / bedKi;
	#endif
  }

  #if defined(Heater1Pin) && (Heater1Pin > -1) 
    SET_OUTPUT(Heater1Pin);
  #endif  
  #if defined(Heater2Pin) && (Heater2Pin > -1) 
    SET_OUTPUT(Heater2Pin);
  #endif  
  #if defined(Heater3Pin) && (Heater3Pin > -1) 
    SET_OUTPUT(Heater3Pin);
  #endif  
    #if defined(Heater4Pin) && (Heater4Pin > -1) 
    SET_OUTPUT(Heater4Pin);
  #endif  
    #if defined(Heater5Pin) && (Heater5Pin > -1) 
    SET_OUTPUT(Heater5Pin);
  #endif  
  #if defined(HeaterBedPin) && (HeaterBedPin > -1) 
    SET_OUTPUT(HeaterBedPin);
  #endif  

  ADCSRA = 1<<ADEN | 1<<ADSC | 1<<ADIF | 0x07;
  DIDR0 = 0;
  #ifdef DIDR2
    DIDR2 = 0;
  #endif
  #if defined(Temp1Pin) && (Temp1Pin > -1)
    #if Temp1Pin < 8
       DIDR0 |= 1 << Temp1Pin; 
    #else
       DIDR2 |= 1<<(Temp1Pin - 8); 
    #endif
  #endif
  #if defined(Temp2Pin) && (Temp2Pin > -1)
    #if Temp2Pin < 8
       DIDR0 |= 1<<Temp2Pin; 
    #else
       DIDR2 |= 1<<(Temp2Pin - 8); 
    #endif
  #endif
  #if defined(Temp4Pin) && (Temp4Pin > -1)
    #if Temp4Pin < 8
       DIDR0 |= 1 << Temp4Pin; 
    #else
       DIDR2 |= 1<<(Temp4Pin - 8); 
    #endif
  #endif
  #if defined(Temp5Pin) && (Temp5Pin > -1)
    #if Temp5Pin < 8
       DIDR0 |= 1 << Temp4Pin; 
    #else
       DIDR2 |= 1<<(Temp5Pin - 8); 
    #endif
  #endif
  #if defined(Temp6Pin) && (Temp6Pin > -1)
    #if Temp6Pin < 8
       DIDR0 |= 1 << Temp6Pin; 
    #else
       DIDR2 |= 1<<(Temp6Pin - 8); 
    #endif
  #endif
  #if defined(Temp7Pin) && (Temp7Pin > -1)
    #if Temp7Pin < 8
       DIDR0 |= 1 << Temp7Pin; 
    #else
       DIDR2 |= 1<<(Temp7Pin - 8); 
    #endif
  #endif
  #if defined(Temp3Pin) && (Temp3Pin > -1)
    #if Temp3Pin < 8
       DIDR0 |= 1<<Temp3Pin; 
    #else
       DIDR2 |= 1<<(Temp3Pin - 8); 
    #endif
  #endif
  
  OCR0B = 128;
  TIMSK0 |= (1<<OCIE0B);  
  
  delay(250);

#ifdef Heater1MinTemp
  minttemp[0] = Heater1MinTemp;
  while(analog2temp(minttemp_raw[0], 0) < Heater1MinTemp) 
  {
#if HEATER_0_RAW_LO_TEMP < HEATER_0_RAW_HI_TEMP
    minttemp_raw[0] += OVERSAMPLENR;
#else
    minttemp_raw[0] -= OVERSAMPLENR;
#endif
  }
#endif
#ifdef Heater1MaxTemp
  maxttemp[0] = Heater1MaxTemp;
  while(analog2temp(maxttemp_raw[0], 0) > Heater1MaxTemp) 
  {
#if HEATER_0_RAW_LO_TEMP < HEATER_0_RAW_HI_TEMP
    maxttemp_raw[0] -= OVERSAMPLENR;
#else
    maxttemp_raw[0] += OVERSAMPLENR;
#endif
  }
#endif

#if (Nozzles > 1) && defined(Heater2MinTemp)
  minttemp[1] = Heater2MinTemp;
  while(analog2temp(minttemp_raw[1], 1) < Heater2MinTemp) 
  {
#if HEATER_1_RAW_LO_TEMP < HEATER_1_RAW_HI_TEMP
    minttemp_raw[1] += OVERSAMPLENR;
#else
    minttemp_raw[1] -= OVERSAMPLENR;
#endif
  }
#endif
#if (Nozzles > 1) && defined(Heater2MaxTemp)
  maxttemp[1] = Heater2MaxTemp;
  while(analog2temp(maxttemp_raw[1], 1) > Heater2MaxTemp) 
  {
#if HEATER_1_RAW_LO_TEMP < HEATER_1_RAW_HI_TEMP
    maxttemp_raw[1] -= OVERSAMPLENR;
#else
    maxttemp_raw[1] += OVERSAMPLENR;
#endif
  }
#endif

#if (Nozzles > 2) && defined(Heater3MinTemp)
  minttemp[2] = Heater3MinTemp;
  while(analog2temp(minttemp_raw[2], 2) < Heater3MinTemp) 
  {
#if HEATER_2_RAW_LO_TEMP < HEATER_2_RAW_HI_TEMP
    minttemp_raw[2] += OVERSAMPLENR;
#else
    minttemp_raw[2] -= OVERSAMPLENR;
#endif
  }
#endif
#if (Nozzles > 2) && defined(Heater3MaxTemp)
  maxttemp[2] = Heater3MaxTemp;
  while(analog2temp(maxttemp_raw[2], 2) > Heater3MaxTemp) 
  {
#if HEATER_2_RAW_LO_TEMP < HEATER_2_RAW_HI_TEMP
    maxttemp_raw[2] -= OVERSAMPLENR;
#else
    maxttemp_raw[2] += OVERSAMPLENR;
#endif
  }
#endif

#if (Nozzles > 3) && defined(Heater4MinTemp)
  minttemp[3] = Heater4MinTemp;
  while(analog2temp(minttemp_raw[3], 3) < Heater4MinTemp) 
  {
#if HEATER_3_RAW_LO_TEMP < HEATER_3_RAW_HI_TEMP
    minttemp_raw[3] += OVERSAMPLENR;
#else
    minttemp_raw[3] -= OVERSAMPLENR;
#endif
  }
#endif
#if (Nozzles > 3) && defined(Heater4MaxTemp)
  maxttemp[3] = Heater4MaxTemp;
  while(analog2temp(maxttemp_raw[3], 3) > Heater4MaxTemp)
  {
#if HEATER_3_RAW_LO_TEMP < HEATER_3_RAW_HI_TEMP
    maxttemp_raw[3] -= OVERSAMPLENR;
#else
    maxttemp_raw[3] += OVERSAMPLENR;
#endif
  }
#endif

#if (Nozzles > 4) && defined(Heater5MinTemp)
  minttemp[4] = Heater5MinTemp;
  while(analog2temp(minttemp_raw[4], 4) < Heater5MinTemp) 
  {
#if HEATER_4_RAW_LO_TEMP < HEATER_4_RAW_HI_TEMP
    minttemp_raw[4] += OVERSAMPLENR;
#else
    minttemp_raw[4] -= OVERSAMPLENR;
#endif
  }
#endif
#if (Nozzles > 4) && defined(Heater5MaxTemp)
  maxttemp[4] = Heater5MaxTemp;
  while(analog2temp(maxttemp_raw[4], 4) > Heater5MaxTemp) 
  {
#if HEATER_4_RAW_LO_TEMP < HEATER_4_RAW_HI_TEMP
    maxttemp_raw[4] -= OVERSAMPLENR;
#else
    maxttemp_raw[4] += OVERSAMPLENR;
#endif
  }
#endif


#ifdef Heater6MinTemp
#endif
#ifdef BedMaxTemp
  while(analog2tempBed(bed_maxttemp_raw) > BedMaxTemp) 
  {
#if HEATER_BED_RAW_LO_TEMP < HEATER_BED_RAW_HI_TEMP
    bed_maxttemp_raw -= OVERSAMPLENR;
#else
    bed_maxttemp_raw += OVERSAMPLENR;
#endif
  }
#endif
}

void setWatch() 
{  
#ifdef WATCH_TEMP_PERIOD
  for (int e = 0; e < Nozzles; e++)
  {
    if(degHotend(e) < degTargetHotend(e) - (WATCH_TEMP_INCREASE * 2))
    {
      watch_start_temp[e] = degHotend(e);
      watchmillis[e] = millis();
    } 
  }
#endif 
}


void disable_heater()
{
  for(int i=0;i<Nozzles;i++)
    setTargetHotend(0,i);
  setTargetBed(0);
  #if defined(Temp1Pin) && Temp1Pin > -1
  target_temperature[0]=0;
  soft_pwm[0]=0;
   #if defined(Heater1Pin) && Heater1Pin > -1  
     WRITE(Heater1Pin,LOW);
   #endif
  #endif
     
  #if defined(Temp2Pin) && Temp2Pin > -1
    target_temperature[1]=0;
    soft_pwm[1]=0;
    #if defined(Heater2Pin) && Heater2Pin > -1 
      WRITE(Heater2Pin,LOW);
    #endif
  #endif
      
  #if defined(Temp4Pin) && Temp4Pin > -1
    target_temperature[2]=0;
    soft_pwm[2]=0;
    #if defined(Heater3Pin) && Heater3Pin > -1  
      WRITE(Heater3Pin,LOW);
    #endif
  #endif 
      
  #if defined(Temp5Pin) && Temp5Pin > -1
    target_temperature[3]=0;
    soft_pwm[3]=0;
    #if defined(Heater4Pin) && Heater4Pin > -1  
      WRITE(Heater4Pin,LOW);
    #endif
  #endif 
      
  #if defined(Temp6Pin) && Temp6Pin > -1
    target_temperature[4]=0;
    soft_pwm[4]=0;
    #if defined(Heater5Pin) && Heater5Pin > -1
      WRITE(Heater5Pin,LOW);
    #endif
  #endif 
      
  #if defined(Temp3Pin) && Temp3Pin > -1
    target_temperature_bed=0;
    soft_pwm_bed=0;
    #if defined(HeaterBedPin) && HeaterBedPin > -1
      WRITE(HeaterBedPin,LOW);
    #endif
  #endif
}

void max_temp_error(uint8_t e) 
{
  disable_heater();
  if(IsStopped() == false) 
  {
    SERIAL_ERROR_START;
    SERIAL_ERRORLN((int)e);
    SERIAL_ERRORLNPGM(": Extruder switched off. MAXTEMP triggered !");
  }
  #ifndef BOGUS_TEMPERATURE_FAILSAFE_OVERRIDE
  Stop();
  #endif
}

void min_temp_error(uint8_t e) 
{
  disable_heater();
  if(IsStopped() == false) 
  {
    SERIAL_ERROR_START;
    SERIAL_ERRORLN((int)e);
    SERIAL_ERRORLNPGM(": Extruder switched off. MINTEMP triggered !");
  }
  #ifndef BOGUS_TEMPERATURE_FAILSAFE_OVERRIDE
  Stop();
  #endif
}

void bed_max_temp_error(void) 
{
#if HeaterBedPin > -1
  WRITE(HeaterBedPin, 0);
#endif
  if(IsStopped() == false) 
  {
    SERIAL_ERROR_START;
    SERIAL_ERRORLNPGM("Temperature heated bed switched off. MAXTEMP triggered !!");
  }
  #ifndef BOGUS_TEMPERATURE_FAILSAFE_OVERRIDE
  Stop();
  #endif
}

ISR(TIMER0_COMPB_vect)
{
  static unsigned char temp_count = 0;
  static unsigned long raw_temp_0_value = 0;
  static unsigned long raw_temp_1_value = 0;
  static unsigned long raw_temp_2_value = 0;
  static unsigned long raw_temp_3_value = 0;  
  static unsigned long raw_temp_4_value = 0;
  static unsigned long raw_temp_LC_value = 0;
  static unsigned long raw_temp_bed_value = 0;
  static unsigned char temp_state = 0;
  static unsigned char pwm_count = (1 << SOFT_PWM_SCALE);
  static unsigned char soft_pwm_0;
  #if Nozzles > 1
  static unsigned char soft_pwm_1;
  #endif
  #if Nozzles > 2
  static unsigned char soft_pwm_2;
  #endif
  #if Nozzles > 3
  static unsigned char soft_pwm_3;
  #endif
  #if Nozzles > 4
  static unsigned char soft_pwm_4;
  #endif
  #if HeaterBedPin > -1
  static unsigned char soft_pwm_b;
  #endif
  
  if(pwm_count == 0)
  {
    soft_pwm_0 = soft_pwm[0];
    if(soft_pwm_0 > 0) WRITE(Heater1Pin,1);
    #if Nozzles > 1
    soft_pwm_1 = soft_pwm[1];
    if(soft_pwm_1 > 0) WRITE(Heater2Pin,1);
    #endif
    #if Nozzles > 2
    soft_pwm_2 = soft_pwm[2];
    if(soft_pwm_2 > 0) WRITE(Heater3Pin,1);
    #endif
    #if Nozzles > 3
    soft_pwm_3 = soft_pwm[3];
    if(soft_pwm_3 > 0) WRITE(Heater4Pin,1);
    #endif
    #if Nozzles > 4
    soft_pwm_4 = soft_pwm[4];
    if(soft_pwm_4 > 0) WRITE(Heater5Pin,1);
    #endif
    #if defined(HeaterBedPin) && HeaterBedPin > -1
    soft_pwm_b = soft_pwm_bed;
    if(soft_pwm_b > 0) WRITE(HeaterBedPin,1);
    #endif
  }
  if(soft_pwm_0 <= pwm_count) WRITE(Heater1Pin,0);
  #if Nozzles > 1
  if(soft_pwm_1 <= pwm_count) WRITE(Heater2Pin,0);
  #endif
  #if Nozzles > 2
  if(soft_pwm_2 <= pwm_count) WRITE(Heater3Pin,0);
  #endif
  #if Nozzles > 3
  if(soft_pwm_3 <= pwm_count) WRITE(Heater4Pin,0);
  #endif
  #if Nozzles > 4
  if(soft_pwm_4 <= pwm_count) WRITE(Heater5Pin,0);
  #endif
  #if defined(HeaterBedPin) && HeaterBedPin > -1
  if(soft_pwm_b <= pwm_count) WRITE(HeaterBedPin,0);
  #endif
  
  pwm_count += (1 << SOFT_PWM_SCALE);

  pwm_count &= 0x7f;
  
  switch(temp_state) 
  {
    case 0:
      #if defined(Temp1Pin) && (Temp1Pin > -1)
        #if Temp1Pin > 7
          ADCSRB = 1<<MUX5;
        #else
          ADCSRB = 0;
        #endif
        ADMUX = ((1 << REFS0) | (Temp1Pin & 0x07));
        ADCSRA |= 1<<ADSC;
      #endif
      temp_state = 1;
      break;
    case 1:
      #if defined(Temp1Pin) && (Temp1Pin > -1)
        raw_temp_0_value += ADC;
      #endif
      #ifdef HEATER_0_USES_MAX6675
        raw_temp_0_value = read_max6675();
      #endif
      temp_state = 2;
      break;
    case 2:
      #if defined(Temp3Pin) && (Temp3Pin > -1)
        #if Temp3Pin > 7
          ADCSRB = 1<<MUX5;
        #else
          ADCSRB = 0;
        #endif
        ADMUX = ((1 << REFS0) | (Temp3Pin & 0x07));
        ADCSRA |= 1<<ADSC;
      #endif
      temp_state = 3;
      break;
    case 3:
      #if defined(Temp3Pin) && (Temp3Pin > -1)
        raw_temp_bed_value += ADC;
      #endif
      temp_state = 4;
      break;
    case 4:
      #if defined(Temp2Pin) && (Temp2Pin > -1)
        #if Temp2Pin > 7
          ADCSRB = 1<<MUX5;
        #else
          ADCSRB = 0;
        #endif
        ADMUX = ((1 << REFS0) | (Temp2Pin & 0x07));
        ADCSRA |= 1<<ADSC;
      #endif
      temp_state = 5;
      break;
    case 5:
      #if defined(Temp2Pin) && (Temp2Pin > -1)
        raw_temp_1_value += ADC;
      #endif
      temp_state = 6;
      break;
    case 6:
      #if defined(Temp4Pin) && (Temp4Pin > -1)
        #if Temp4Pin > 7
          ADCSRB = 1<<MUX5;
        #else
          ADCSRB = 0;
        #endif
        ADMUX = ((1 << REFS0) | (Temp4Pin & 0x07));
        ADCSRA |= 1<<ADSC;
      #endif
      temp_state = 7;
      break;
    case 7:
      #if defined(Temp4Pin) && (Temp4Pin > -1)
        raw_temp_2_value += ADC;
      #endif
      temp_state = 8;
      break;
    case 8:
      #if defined(Temp4Pin) && (Temp4Pin > -1)
        #if Temp4Pin > 7
          ADCSRB = 1<<MUX5;
        #else
          ADCSRB = 0;
        #endif
        ADMUX = ((1 << REFS0) | (Temp4Pin & 0x07));
        ADCSRA |= 1<<ADSC;
      #endif
      temp_state = 9;
      break;
    case 9:
      #if defined(Temp4Pin) && (Temp4Pin > -1)
        raw_temp_3_value += ADC;
      #endif
      temp_state = 10;
      break;
    case 10:
      #if defined(Temp6Pin) && (Temp6Pin > -1)
        #if Temp6Pin > 7
          ADCSRB = 1<<MUX5;
        #else
          ADCSRB = 0;
        #endif
        ADMUX = ((1 << REFS0) | (Temp6Pin & 0x07));
        ADCSRA |= 1<<ADSC;
      #endif
      temp_state = 11;
      break;
    case 11:
      #if defined(Temp6Pin) && (Temp6Pin > -1)
        raw_temp_4_value += ADC;
		temp_state = 12;
      #endif
      break;
    case 12:
      #if defined(Temp7Pin) && (Temp7Pin > -1)
        #if Temp7Pin > 7
          ADCSRB = 1<<MUX5;
        #else
          ADCSRB = 0;
        #endif
        ADMUX = ((1 << REFS0) | (Temp7Pin & 0x07));
        ADCSRA |= 1<<ADSC;
      #endif
      temp_state = 13;
      break;
    case 13:
      #if defined(Temp7Pin) && (Temp7Pin > -1)
        raw_temp_LC_value += ADC;
      #endif
      temp_state = 0;
      temp_count++;
      break;
  }
    
  if(temp_count >= 16)
  {
    if (!temp_meas_ready)
    {
      current_temperature_raw[0] = raw_temp_0_value;
#if Nozzles > 1
      current_temperature_raw[1] = raw_temp_1_value;
#endif
#if Nozzles > 2
      current_temperature_raw[2] = raw_temp_2_value;
#endif
#if Nozzles > 3
      current_temperature_raw[3] = raw_temp_3_value;
#endif
#if Nozzles > 4
      current_temperature_raw[4] = raw_temp_4_value;
#endif
	  current_temperature_LC_raw = raw_temp_LC_value;
      current_temperature_bed_raw = raw_temp_bed_value;
    }
    
    temp_meas_ready = true;
    temp_count = 0;
    raw_temp_0_value = 0;
    raw_temp_1_value = 0;
    raw_temp_2_value = 0;
    raw_temp_3_value = 0;
    raw_temp_4_value = 0;
	raw_temp_LC_value = 0;
    raw_temp_bed_value = 0;

#if HEATER_0_RAW_LO_TEMP > HEATER_0_RAW_HI_TEMP
    if(current_temperature_raw[0] <= maxttemp_raw[0]) 
	{
#else
    if(current_temperature_raw[0] >= maxttemp_raw[0]) 
	{
#endif
        max_temp_error(0);
    }
#if HEATER_0_RAW_LO_TEMP > HEATER_0_RAW_HI_TEMP
    if(current_temperature_raw[0] >= minttemp_raw[0]) 
	{
#else
    if(current_temperature_raw[0] <= minttemp_raw[0]) 
	{
#endif
        min_temp_error(0);
    }
#if Nozzles > 1
#if HEATER_1_RAW_LO_TEMP > HEATER_1_RAW_HI_TEMP
    if(current_temperature_raw[1] <= maxttemp_raw[1]) 
	{
#else
    if(current_temperature_raw[1] >= maxttemp_raw[1]) 
	{
#endif
        max_temp_error(1);
    }
#if HEATER_1_RAW_LO_TEMP > HEATER_1_RAW_HI_TEMP
    if(current_temperature_raw[1] >= minttemp_raw[1]) 
	{
#else
    if(current_temperature_raw[1] <= minttemp_raw[1]) 
	{
#endif
        min_temp_error(1);
    }
#endif
#if Nozzles > 2
#if HEATER_2_RAW_LO_TEMP > HEATER_2_RAW_HI_TEMP
    if(current_temperature_raw[2] <= maxttemp_raw[2]) 
	{
#else
    if(current_temperature_raw[2] >= maxttemp_raw[2]) 
	{
#endif
        max_temp_error(2);
    }
#if HEATER_2_RAW_LO_TEMP > HEATER_2_RAW_HI_TEMP
    if(current_temperature_raw[2] >= minttemp_raw[2]) 
	{
#else
    if(current_temperature_raw[2] <= minttemp_raw[2])
	{
#endif
        min_temp_error(2);
    }
#endif
#if Nozzles > 3
#if HEATER_3_RAW_LO_TEMP > HEATER_3_RAW_HI_TEMP
    if(current_temperature_raw[3] <= maxttemp_raw[3]) 
	{
#else
    if(current_temperature_raw[3] >= maxttemp_raw[3]) 
	{
#endif
        max_temp_error(3);
    }
#if HEATER_3_RAW_LO_TEMP > HEATER_3_RAW_HI_TEMP
    if(current_temperature_raw[3] >= minttemp_raw[3]) 
	{
#else
    if(current_temperature_raw[3] <= minttemp_raw[3]) 
	{
#endif
        min_temp_error(3);
    }
#endif
#if Nozzles > 4
#if HEATER_4_RAW_LO_TEMP > HEATER_4_RAW_HI_TEMP
    if(current_temperature_raw[4] <= maxttemp_raw[4]) 
	{
#else
    if(current_temperature_raw[4] >= maxttemp_raw[4]) 
	{
#endif
        max_temp_error(4);
    }
#if HEATER_4_RAW_LO_TEMP > HEATER_4_RAW_HI_TEMP
    if(current_temperature_raw[4] >= minttemp_raw[4]) 
	{
#else
    if(current_temperature_raw[4] <= minttemp_raw[4]) 
	{
#endif
        min_temp_error(4);
    }
#endif
  
#if defined(BedMaxTemp) && (Temp3 != 0)
# if HEATER_BED_RAW_LO_TEMP > HEATER_BED_RAW_HI_TEMP
    if(current_temperature_bed_raw <= bed_maxttemp_raw) 
	{
#else
    if(current_temperature_bed_raw >= bed_maxttemp_raw)
	{
#endif
       target_temperature_bed = 0;
       bed_max_temp_error();
    }
#endif
  }  
}

#ifdef PidTemp

float scalePID_i(float i)
{
	return i*Pid_dT;
}

float unscalePID_i(float i)
{
	return i/Pid_dT;
}

float scalePID_d(float d)
{
    return d/Pid_dT;
}

float unscalePID_d(float d)
{
	return d*Pid_dT;
}

#endif


